# Copyright (c) 2019 Kevin Weiss, for HAW Hamburg  <kevin.weiss@haw-hamburg.de>
#
# This file is subject to the terms and conditions of the MIT License. See the
# file LICENSE in the top level directory for more details.
# SPDX-License-Identifier:    MIT
"""Interfaces for PHiLIP in philip_pal
This handles the basic protocol interface to Philip and extended helpers for
Philip.
"""
import logging
import errno
import os
import json
import csv
import time
from ast import literal_eval
from pathlib import Path
from .base_device import BaseDevice


class PhilipBaseIf():
    """Handles basic functions and commands for memory map interface

    Args:
        (*args, **kwargs) -> See base_device for documentation of args
    """
    RESULT_SUCCESS = 'Success'
    RESULT_ERROR = 'Error'
    RESULT_TIMEOUT = 'Timeout'

    def __init__(self, *args, **kwargs):
        self.dev = BaseDevice(*args, **kwargs)

    @staticmethod
    def _error_msg(data):
        if data not in errno.errorcode:
            return "Unknown Error[{}]".format(data)
        s_errcode = errno.errorcode[data]
        s_errmsg = os.strerror(data)
        return "{}-{} [{}]".format(s_errcode, s_errmsg, data)

    @staticmethod
    def _bytes_to_int(data_list):
        """
        >>> PhilipBaseIf._bytes_to_int([0,1])
        256
        """
        val = 0
        try:
            for index, num in enumerate(data_list):
                if num > 255 or num < 0:
                    raise ValueError
                val += num << (index * 8)
        except (ValueError, TypeError):
            return data_list
        return val

    @staticmethod
    def _try_data_to_int(cmd_info):
        if 'data' in cmd_info:
            cmd_info['data'] = PhilipBaseIf._bytes_to_int(cmd_info['data'])

    def _send_cmd(self, send_cmd, timeout, end_key='result',):
        self.dev.writeline(send_cmd)
        cmd_info = {}
        decode_error = False
        while end_key not in cmd_info:
            line = self.dev.readline(timeout)
            try:
                cmd_info.update(json.loads(line))
            except json.decoder.JSONDecodeError:
                decode_error = True
        if decode_error:
            raise json.decoder.JSONDecodeError
        return cmd_info

    def send_and_parse_cmd(self, send_cmd, to_byte_array=False, timeout=None):
        """Returns a dictionary with information from the event
        Args:
            send_cmd(str): The command to write to the device
            to_byte_array: If True and data is bytes leave it as an array
            timeout: Optional timeout value for command specific timeouts
        Returns:
            dict:
            The return hold dict values in the following keys::
            msg - The message from the response, only used for information.
            cmd - The command sent, used to track what has occured.
            data - Parsed information of the data requested.
            result - Either success, error or timeout.
        """

        cmd_info = {'cmd': send_cmd}
        try:
            cmd_info.update(self._send_cmd(send_cmd, timeout=timeout))
            if to_byte_array is False:
                if 'data' in cmd_info:
                    self._try_data_to_int(cmd_info)
        except json.decoder.JSONDecodeError:
            cmd_info['result'] = self.RESULT_ERROR
            cmd_info['msg'] = "Failed to parse response"
        except TimeoutError:
            cmd_info['result'] = self.RESULT_TIMEOUT
        else:
            if cmd_info['result'] == 0:
                cmd_info['result'] = self.RESULT_SUCCESS
            else:
                cmd_info['data'] = cmd_info['result']
                cmd_info['result'] = self.RESULT_ERROR
                cmd_info['msg'] = self._error_msg(cmd_info['data'])
        return cmd_info

    def read_bytes(self, index, size=1, to_byte_array=False, timeout=None):
        """Reads bytes in the register map
        Args:
            index: Index of the memory map (address or offset of bytes)
            size: The amount of bytes to read
            to_byte_array: If True and data is bytes leave it as an array
            timeout: Optional timeout value for command specific timeouts
        Returns:
            see send_and_parse_cmd()
        """
        return self.send_and_parse_cmd('rr {} {}'.format(index, size),
                                       to_byte_array, timeout)

    def read_bits(self, index, offset, bit_amount, timeout=None):
        """Read specific bits in the register map
        Args:
            index: Index of the memory map (address or offset of bytes)
            offset: The bit offset for the bitfield
            bit_amount: The amount of bits within the bitfield
            timeout: Optional timeout value for command specific timeouts
        Returns:
            see send_and_parse_cmd()
        """
        bit_amount = int(bit_amount)
        offset = int(offset)
        index = int(index)
        bytes_to_read = int((bit_amount - 1 + offset)/8 + 1)
        bit_mask = (2 ** bit_amount) - 1
        cmd_info = self.read_bytes(index, bytes_to_read, timeout=timeout)
        if cmd_info['result'] == self.RESULT_SUCCESS:
            cmd_info['cmd'] = ['read_bits({},{},{})'.format(index, offset,
                                                            bit_amount),
                               cmd_info['cmd']]
            cmd_info['data'] = cmd_info['data'] >> offset
            cmd_info['data'] = cmd_info['data'] & bit_mask
        return cmd_info

    @staticmethod
    def _write_byte_arg_to_string(data, size):
        ret_str = ''
        if isinstance(data, list):
            for data_byte in data:
                ret_str += ' {}'.format(literal_eval(str(data_byte)))
        else:
            for i in range(0, size):
                ret_str += ' {}'.format((int(data) >> ((i) * 8)) & 0xFF)
        return ret_str

    def write_bytes(self, index, data, size=4, timeout=None):
        """Writes bytes in the register map
        Args:
            index: Index of the memory map (address or offset of bytes)
            data: The data to be converted to bytes then written to the map
            size: The amount of bytes to write, used if data is not a list
            timeout: Optional timeout value for command specific timeouts
        Returns:
            see send_and_parse_cmd()
        """
        cmd = "wr {}{}".format(index,
                               self._write_byte_arg_to_string(data, size))
        return self.send_and_parse_cmd(cmd, timeout=timeout)

# pylint: disable=R0913
    def write_bits(self, index, offset, bit_amount, data, timeout=None):
        """Modifies specific bits in the register map
        Args:
            index: Index of the memory map (address or offset of bytes)
            offset: The bit offset for the bitfield
            bit_amount: The amount of bits within the bitfield
            data: The data to be converted to bytes then written to the map
            timeout: Optional timeout value for command specific timeouts
        Returns:
            see send_and_parse_cmd()
        """
        cmds = ['write_bits({},{},{},{})'.format(index, offset,
                                                 bit_amount, data)]
        bit_amount = int(bit_amount)
        offset = int(offset)
        index = int(index)
        bytes_to_read = int((bit_amount - 1 + offset)/8 + 1)
        cmd_info = self.read_bytes(index, bytes_to_read, timeout=timeout)
        if cmd_info['result'] != self.RESULT_SUCCESS:
            return cmd_info
        cmds.append(cmd_info['cmd'])
        bit_mask = int((2 ** bit_amount) - 1)
        bit_mask = bit_mask << offset
        cmd_info['data'] = cmd_info['data'] & (~bit_mask)
        shifted_data = cmd_info['data'] | ((data << offset) & bit_mask)
        cmd_info = self.write_bytes(index, shifted_data,
                                    bytes_to_read, timeout=timeout)
        cmds.append(cmd_info['cmd'])
        if cmd_info['result'] == self.RESULT_SUCCESS:
            cmd_info['cmd'] = 'write_bits {} {} {} {}'.format(index, offset,
                                                              bit_amount, data)
        cmd_info['cmd'] = cmds
        return cmd_info

    def execute_changes(self, timeout=None):
        """Executes or commits device configuration changes

        This will cause any changes in configuration to be applied.  For many
        periphs the .mode.init must be set to 0 for the periph to reinitialize.

        Args:
            timeout: Optional timeout value for command specific timeouts
        Returns:
            see send_and_parse_cmd()
        """
        return self.send_and_parse_cmd("ex", timeout=timeout)

    def reset_mcu(self, timeout=None):
        """Resets the device

        Args:
            timeout: Optional timeout value for command specific timeouts
        Returns:
            see send_and_parse_cmd()
        """
        return self.send_and_parse_cmd("mcu_rst", timeout=timeout)

    def get_version(self, timeout=None):
        """Get the version of the interface/memory map that is being used

        Args:
            timeout: Optional timeout value for command specific timeouts
        Returns:
            returns a dict with the {version: x.x.x}
        """
        return self.send_and_parse_cmd("version", timeout=timeout)

    def get_mem_map(self, index, timeout=None):
        """Gets the memory map information from the device

        Args:
            index: The index of the memory map record
            timeout: Optional timeout value for command specific timeouts
        Returns:
            Returns a dict with a memory map map record
        """
        return self.send_and_parse_cmd("mm {}".format(index), timeout=timeout)

    def get_mem_map_size(self, timeout=None):
        """Gets the amount of records in the device memory map

        Args:
            timeout: Optional timeout value for command specific timeouts
        Returns:
            returns a dict with the {data: number of records}
        """
        return self.send_and_parse_cmd("mm_size", timeout=timeout)


class PhilipExtIf(PhilipBaseIf):
    """Handles extended commands to PHiLIP device

    This handles extended commands that involve a combination of basic
    commands.  It also handles getting the memory map of the philip device.

    Args:
        map_path - keyword arg that allows specifying a path for the
                   memory map
        use_dev_map - keyword arg that scans the device for the internal
                      memory map (it may take longer and not have as much
                      information)
        (*args, **kwargs) -> See base_device for documentation of args
    """

    def __init__(self, *args, **kwargs):
        map_path = ''
        use_dev_map = False
        if 'map_path' in kwargs:
            map_path = kwargs.pop('map_path')
        if 'use_dev_map' in kwargs:
            use_dev_map = kwargs.pop('use_dev_map')

        super().__init__(*args, **kwargs)

        self.if_version = self.get_version()['version']

        if map_path != '':
            self.mem_map = self.import_mm_from_csv(map_path)
        elif use_dev_map:
            self.mem_map = self.scan_dev_map()
        else:
            self.mem_map = self.mm_from_version(self.if_version)

    def scan_dev_map(self):
        """Scans the device for memory map information

        Returns:
            Basic memory map from the device
        """
        prim_types = {'uint8_t': 1, 'int8_t': 1, 'uint16_t': 2, 'int16_t': 2,
                      'uint32_t': 4, 'int32_t': 4, 'uint64_t': 8, 'int64_t': 8,
                      'char': 1, 'float': 4, 'double': 8}
        res = self.get_mem_map_size()
        if res['result'] == self.RESULT_ERROR:
            raise NotImplementedError("Cannot find memory map on device")
        mm_size = res['data']
        mem_map = {}
        for map_index in range(0, mm_size):
            res = self.get_mem_map(map_index)
            res['type_size'] = prim_types[res['type']]
            if res['array_size'] == 0:
                res['array_size'] = ''
                res['total_size'] = ''
            else:
                res['total_size'] = res['type_size'] * res['array_size']
            if res['bits'] == 0:
                res['bits'] = ''
                res['bit_offset'] = ''
            res.pop('cmd')
            res.pop('result')
            mem_map[res['name']] = res
        return mem_map

    @staticmethod
    def _try_parse_int_list(list_int):
        for i, val in enumerate(list_int):
            try:
                list_int[i] = literal_eval(str(val))
            except (ValueError, TypeError, SyntaxError):
                pass

    @staticmethod
    def mm_from_version(version_str):
        """Gets the saved memory map given the version
        Args:
            version_str: The version string (x.x.x)
        Returns:
            Versioned memory map
        """
        version_str = version_str.replace('.', '_')
        rel_path = '/mem_map/PHiLIP_map_t_{}.csv'.format(version_str)
        version_path = str(Path(__file__).parents[0]) + rel_path
        return PhilipExtIf.import_mm_from_csv(version_path)

    @staticmethod
    def import_mm_from_csv(path):
        """Imports a memory map csv file

        Returns:
            memory map from the CSV"""
        mem_map = {}
        with open(path) as csvfile:
            rows = list(csv.reader(csvfile, quotechar="'"))
            for row in range(1, len(rows)):
                PhilipExtIf._try_parse_int_list(rows[row])
                cmd = dict(zip(rows[0], rows[row]))
                mem_map[rows[row][rows[0].index('name')]] = cmd
                logging.debug("Imported command: %r", cmd)
        return mem_map

    @staticmethod
    def _parse_array(data, type_size):
        parsed_data = []
        try:
            elements = int(len(data)/type_size)
            for i in range(0, elements):
                num = int.from_bytes(data[i*type_size:(i+1)*type_size],
                                     byteorder='little')
                parsed_data.append(num)
        except (ValueError, TypeError):
            return data
        return parsed_data

# pylint: disable=R0913
    def read_reg(self, cmd_name, offset=0, size=None,
                 to_byte_array=False, timeout=None):
        """Read a register defined by the memory map

        Args:
            cmd_name(str): The name of the register to read
            offset(int): The number of elements to offset in an array
            size(int): The number of elements to read in an array
            to_byte_array: If True and data is bytes leave it as an array
            timeout: Optional timeout value for command specific timeouts
        Returns:
            see send_and_parse_cmd()
        """
        cmd = self.mem_map[cmd_name]
        response = None
        if size is None:
            if cmd['total_size'] == '':
                size = cmd['type_size']
            else:
                size = cmd['total_size']
        else:
            if cmd['total_size'] != '':
                size = int(cmd['type_size']) * int(size)
        if cmd['bits'] != '':
            response = self.read_bits(cmd['offset'],
                                      cmd['bit_offset'],
                                      cmd['bits'],
                                      timeout=timeout)
        elif cmd['total_size'] != '':
            offset = int(offset)
            offset *= int(cmd['type_size'])
            offset += int(cmd['offset'])
            response = self.read_bytes(offset, size, True, timeout)
            response['data'] = self._parse_array(response['data'],
                                                 cmd['type_size'])
        else:
            response = self.read_bytes(cmd['offset'], size, to_byte_array)
        response['cmd'] = ['read_reg({},{},{})'.format(cmd_name, offset, size),
                           response['cmd']]
        return response

    def write_reg(self, cmd_name, data, offset=0, timeout=None):
        """Writes a register defined by the memory map

        Args:
            cmd_name(str): The name of the register to write
            data: The data to write to the register
            offset(int): The number of elements to offset in an array
            timeout: Optional timeout value for command specific timeouts
        Returns:
            see send_and_parse_cmd()
        """
        cmd = self.mem_map[cmd_name]
        data = literal_eval(str(data))
        response = None
        if cmd['bits'] != '':
            response = self.write_bits(cmd['offset'],
                                       cmd['bit_offset'],
                                       cmd['bits'], data)
        elif cmd['total_size'] != '':
            offset = int(offset)
            offset *= int(cmd['type_size'])
            offset += int(cmd['offset'])
            response = self.write_bytes(offset, data, timeout=timeout)
        else:
            response = self.write_bytes(cmd['offset'], data,
                                        int(cmd['type_size']), timeout=timeout)
        response['cmd'] = ['write_reg({},{},{})'.format(cmd_name, data,
                                                        offset),
                           response['cmd']]
        return response

    def read_struct(self, cmd_name, data_has_name=True, timeout=None):
        """Reads a set of registers defined by the memory map

        Args:
            cmd_name(str): The name if the structure to read
            data_has_cmd: Include the record name in the data
            timeout: Optional timeout value for command specific timeouts
        Returns:
            see send_and_parse_cmd()
        """
        response = []
        for cmd in self.mem_map.keys():
            # Supress unused variable warning

            if cmd.startswith(cmd_name):
                response.append(self.read_reg(cmd, timeout=timeout))
                if data_has_name:
                    response[-1]['data'] = {cmd: response[-1]['data']}
        return response

    def write_and_execute(self, cmd_name, data, timeout=None):
        """Writes the register and the init for the struct and executes changes

        Args:
            cmd_name(str): The name of the register to write
            data: The data to write to the register
            timeout: Optional timeout value for command specific timeouts
        Returns:
            see send_and_parse_cmd()
        """
        response = []
        # could be better, now it assumes everying stuct has mode.init
        if '.' in cmd_name:
            struct_name = cmd_name.split('.')[0]
            struct_name = struct_name + '.mode.init'
            response.append(self.write_reg(struct_name, 0, timeout=timeout))
        response.append(self.write_reg(cmd_name, data, timeout=timeout))
        response.append(self.execute_changes(timeout=timeout))
        return response

    def dut_reset(self, reset_time=0.3, timeout=None):
        """Provides a reset to the dut

        Toggles the dut reset pin for the reset time.

        Args:
            reset_time: The duration the dut is put in reset
            timeout: Optional timeout value for command specific timeouts
        Returns:
            see send_and_parse_cmd()
        """
        reset_time = float(reset_time)
        response = []
        response.extend(self.write_and_execute('sys.mode.dut_rst', 1,
                                               timeout=timeout))
        time.sleep(reset_time)
        response.extend(self.write_and_execute('sys.mode.dut_rst', 0,
                                               timeout=timeout))
        return response

    def read_trace(self, to_ns=False):
        """Reads event trace from the dut

        Returns:
            see send_and_parse_cmd()
            formatted to return the time sorted events
        """
        clock_speed = self.read_reg('sys.sys_clk')['data']
        trace = []
        response = {"cmd": "read_trace()", "result": self.RESULT_SUCCESS}

        for i in range(0, self.mem_map['trace.tick']['array_size']):
            trace_event = {}
            # should be time in seconds
            # read only one at a time since the buffer can overflow
            trace_tick_div = self.read_reg('trace.tick_div', i, 1)['data'][0]
            trace_source = self.read_reg('trace.source', i, 1)['data'][0]
            trace_tick = self.read_reg('trace.tick', i, 1)['data'][0]
            trace_value = self.read_reg('trace.value', i, 1)['data'][0]
            total_tick = trace_tick << trace_tick_div
            time_sec = float(total_tick) / clock_speed
            if to_ns:
                trace_event['time'] = int(time_sec * 10000000000)
            else:
                trace_event['time'] = round(time_sec, 10)
            if trace_source == 1:
                trace_event['source'] = 'DEBUG0'
            elif trace_source == 2:
                trace_event['source'] = 'DEBUG1'
            elif trace_source == 3:
                trace_event['source'] = 'DEBUG2'
            else:
                trace_event['source'] = trace_source
            trace_event['event'] = trace_value
            if trace_source != 0:
                trace.append(trace_event)
        response['data'] = sorted(trace, key=lambda x: x['time'])
        return response



# Copyright (c) 2019 Kevin Weiss, for HAW Hamburg  <kevin.weiss@haw-hamburg.de>
#
# This file is subject to the terms and conditions of the MIT License. See the
# file LICENSE in the top level directory for more details.
# SPDX-License-Identifier:    MIT
"""packet init for PHiLIP PAL
This exposes useful modules in the PHiLIP PAL packet
"""
from .philip_if import PhilipExtIf as Phil

__all__ = ['Phil']



# Copyright (c) 2018 Kevin Weiss, for HAW Hamburg  <kevin.weiss@haw-hamburg.de>
#
# This file is subject to the terms and conditions of the MIT License. See the
# file LICENSE in the top level directory for more details.
# SPDX-License-Identifier:    MIT
"""Abstract Base Device for philip_pal
This module is the abstract device that interfaces to a driver.  It handles the
selection and initialzation of the driver.

Example:
    class ApplicationDevice(BaseDevice):
        def show_example(self):
            self._write('Send Example')
            print(self._read())

    example_use_case = ApplicationDevice(driver_type='serial',
                                         port='my/port/name')
    example_use_case.show_example()
"""
from .serial_driver import SerialDriver


class BaseDevice:
    """Instance for devices to connect and utilize drivers.

    Args:
        driver_type(str): Selects the driver that is used on the devices.
            'serial' uses the standard serial port, all following arguments
            get passed through.
        (*args, **kwargs) -> See selected driver_type for documentation of args
    """

    def __init__(self, *args, **kwargs):
        self._driver = self._driver_from_config(*args, **kwargs)

    def open(self, *args, **kwargs):
        """Opens the device connection."""
        self._driver.open(*args, **kwargs)

    def close(self):
        """Closes the device connection."""
        self._driver.close()

    def readline(self, timeout=None):
        """Reads data from the driver.

        Args:
            timeout: Optional timeout value for command specific timeouts
        Returns:
            str: string of data if success, driver defined error if failed.
        """
        if timeout is None:
            return self._driver.readline()
        return self._driver.readline(float(timeout))

    def writeline(self, line):
        """Writes data to the driver.

        Args:
            line(str): Variable length argument list.
        """
        return self._driver.writeline(line)

    @staticmethod
    def _driver_from_config(*args, **kwargs):
        """Returns driver instance given configuration"""
        if 'driver_type' in kwargs:
            driver_type = kwargs.pop('driver_type')
            if driver_type == 'serial':
                return SerialDriver(*args, **kwargs)
            raise NotImplementedError()
        else:
            return SerialDriver(*args, **kwargs)



# Copyright (c) 2019 Kevin Weiss, for HAW Hamburg  <kevin.weiss@haw-hamburg.de>
#
# This file is subject to the terms and conditions of the MIT License. See the
# file LICENSE in the top level directory for more details.
# SPDX-License-Identifier:    MIT
"""packet init for PHiLIP PAL
This exposes a main function to run the packet from CLI
"""
import argparse
import logging

from .philip_if import PhilipExtIf as Phil


def main():
    """
    Main function to allow direct exec of philip_pal package from CLI
    """
    parser = argparse.ArgumentParser(prog='python3 -m philip_pal')
    parser.add_argument('--verbose', action='store_true',
                        help='Enable more output')
    parser.add_argument('--dut_reset', action='store_true',
                        help='Reset device-under-test (DUT)')
    parser.add_argument('--reset', action='store_true',
                        help='Reset PHiLIP MCU')
    parser.add_argument('--fw', '-f', action='store_true',
                        help='Gets the firmware version')
    parser.add_argument('port', action='store', type=str, nargs='?',
                        help='PHiLIP serial port')
    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.INFO)

    logging.info("Connect to PHiLIP")
    if args.port:
        logging.info(". on port: %s", args.port)
        philip = Phil(args.port)
    else:
        philip = Phil()

    if args.fw:
        rev = philip.read_reg('sys.fw_rev')['data']
        print("{\"version\": \"%r.%r.%r\"}" % (rev[3], rev[2], rev[1]))
    if args.reset:
        logging.info(". reset")
        philip.reset_mcu()
    if args.dut_reset:
        logging.info(". dut reset")
        philip.dut_reset()


main()



# Copyright (c) 2018 Kevin Weiss, for HAW Hamburg  <kevin.weiss@haw-hamburg.de>
#
# This file is subject to the terms and conditions of the MIT License. See the
# file LICENSE in the top level directory for more details.
# SPDX-License-Identifier:    MIT
"""Serial Driver for philip_pal
This module handles generic connection and IO to the serial driver.
"""
import os
import logging
import time
from serial import Serial, serial_for_url, SerialException
from serial.tools import list_ports


class SerialDriver:
    """Serial Port Driver

    Contains all reusable functions for connecting, sending and receiving
    data.  Arguments are passed through to the standard pyserial driver.  The
    defaults are changed.  Also if env variables are defined they get used as
    defaults.  Automatically opens the serial port on initialize.  If nothing
    is specified the port is set to the first available port.

    env_vars:
        PHILIP_BAUD: Baudrate of the philip devices (eg. 115200)
        PHILIP_PORT: Port of the philip device (eg. /dev/ttyACM0)
    Args:
        (*args, **kwargs) -> See pyserial for documentation of args
    """
    DEFAULT_TIMEOUT = 0.5
    DEFAULT_BAUDRATE = 115200

    def __init__(self, *args, **kwargs):
        self._connect(*args, **kwargs)
        # Used to clear the cpu and mcu buffer from startup junk data
        time.sleep(0.05)
        self.writeline('')
        self.readline(0.1)

    def _connect(self, *args, **kwargs):
        if 'timeout' not in kwargs:
            kwargs['timeout'] = self.DEFAULT_TIMEOUT
        if len(args) < 2:
            if 'baudrate' not in kwargs:
                if 'PHILIP_BAUD' in os.environ:
                    kwargs['baudrate'] = os.environ['PHILIP_BAUD']
                else:
                    kwargs['baudrate'] = self.DEFAULT_BAUDRATE
        if len(args) == 0:
            if 'port' not in kwargs:
                if 'PHILIP_PORT' in os.environ:
                    kwargs['port'] = os.environ['PHILIP_PORT']
                else:
                    kwargs['port'] = list_ports.comports()[0][0]
        logging.debug("Serial connection args %r -- %r", args, kwargs)
        try:
            self._dev = Serial(*args, **kwargs)
        except SerialException:
            self._dev = serial_for_url(*args, **kwargs)
        self._args = args
        self._kwargs = kwargs

    def open(self, *args, **kwargs):
        """Open a Serial Connection

        env_vars:
            PHILIP_BAUD: Baudrate of the philip devices (eg. 115200)
            PHILIP_PORT: Port of the philip device (eg. /dev/ttyACM0)
        Args:
            (*args, **kwargs) -> See pyserial for documentation of args
        """
        logging.debug("Opening port")
        self._connect(*args, **kwargs)

    def close(self):
        """Close serial connection

        """
        logging.debug("Closing %s", self._dev.port)
        self._dev.close()

    def readline(self, timeout=None):
        """Read Line from Serial

        Read and decode to utf-8 data from the serial port.  Raises
        TimeoutError exception if an nothing is read.

        Args:
            timeout: Optional timeout value for command specific timeouts
        Returns:
            str: string of utf8 encoded data
        """
        if timeout is None:
            res_bytes = self._dev.readline()
        else:
            default_timeout = self._dev.timeout
            self._dev.timeout = timeout
            res_bytes = self._dev.readline()
            self._dev.timeout = default_timeout
        response = res_bytes.decode("utf-8", errors="replace")
        # Gets rid of leading zeros caused by reset
        response = response.strip('\0')
        if response == "":
            # try to reconnect if timout occurs
            self.close()
            self._connect(*self._args, **self._kwargs)
            raise TimeoutError("Timeout during serial readline")
        logging.debug("Response: %s", response.replace('\n', ''))
        return response

    def writeline(self, line):
        """Write Line to Serial
        Writes line to the serial port and adds a newline and encode to utf-8.

        Args:
            line(str): string or list of bytes to send to the driver.
        """
        # Clear the input buffer in case it junk data go in creating an offset
        self._dev.reset_input_buffer()
        logging.debug("Sending: " + line)
        self._dev.write((line + '\n').encode('utf-8'))



#! /usr/bin/env python3
# Copyright (c) 2019 Kevin Weiss, for HAW Hamburg  <kevin.weiss@haw-hamburg.de>
#
# This file is subject to the terms and conditions of the MIT License. See the
# file LICENSE in the top level directory for more details.
# SPDX-License-Identifier:    MIT

"""
This script handles interfacing to the PHiLIP device.  It parses data and
exposes from the PhilipExtIf class to be run in a shell.

The purpose of this script is allow easy setup and manual usage of PHiLIP.

Usage
-----

```
usage: philip_shell.py  [-h]
                        [--loglevel {debug,info,warning,error,fatal,critical}]
                        [--port PORT]
                        [--filter-data]

optional arguments:
  -h, --help            show this help message and exit
  --loglevel {debug,info,warning,error,fatal,critical}
                        Python logger log level (default: warning)
  --port, -p
                        Specify the serial port
  --data_only, -do
                        Filters data from philip responses to only display
                        what is needed (default: False)
```
"""
import cmd
from json import dumps
import logging
import argparse
import serial.tools.list_ports
try:
    import readline
except ImportError:
    readline = None
try:
    from .philip_if import PhilipExtIf
except ImportError:
    from philip_pal.philip_if import PhilipExtIf


class PhilipShell(cmd.Cmd):
    """Command loop for the PHiLIP interface

    Args:
        port - Serial port for the PHiLIP, if None connection wizard tries to
               connent
        data_only - If true only data prints from command an not the whole
                    response struct
    """
    prompt = 'PHiLIP: '

    def __init__(self, port=None, data_only=True, use_dev_map=False):
        if port is None:
            self.phil = self._connect_wizard(use_dev_map=use_dev_map)
        else:
            self.phil = PhilipExtIf(port, use_dev_map=use_dev_map)
        print("Interface Version: {}".format(self.phil.if_version))
        self.data_only = data_only
        cmd.Cmd.__init__(self)

    @staticmethod
    def _connect_wizard(use_dev_map=False):
        print("Starting PHiLIP shell")
        serial_devices = serial.tools.list_ports.comports()
        if len(serial_devices) == 0:
            raise ConnectionError("Could not find any available devices")
        elif len(serial_devices) == 1:
            print('Connected to {}'.format(serial_devices[0]))
            return PhilipExtIf(port=serial_devices[0][0],
                               use_dev_map=use_dev_map)
        else:
            print('Select a serial port:')
            for i, s_dev in enumerate(serial_devices):
                print("{}: {}".format(i, s_dev))
            s_num = input("Selection(number): ")
            return PhilipExtIf(port=serial_devices[int(s_num)][0],
                               use_dev_map=use_dev_map)

    def preloop(self):
        if readline:
            try:
                readline.read_history_file()
            except IOError:
                pass

    def do_read_reg(self, arg):
        """Read a register defined by the memory map

        Usage:
            read_reg <cmd_name> [offset] [size] [to_byte_array] [timeout]

        Args:
            cmd_name: The name of the register to read
            offset: The number of elements to offset in an array
            size: The number of elements to read in an array
            to_byte_array: If True and data is bytes leave it as an array
            timeout: Optional timeout value for command specific timeouts

        """
        self._print_func_result(self.phil.read_reg, arg)

    def complete_read_reg(self, text, line, begidx, endidx):
        """Completes arg with memory map record names"""
        begidx = begidx
        endidx = endidx
        return self._complete_map(text, line)

    def do_write_reg(self, arg):
        """Writes a register defined by the memory map

        If writing to change a periph configuration the .mode.init bit should
        be set to 0 for reinitialization to occur.

        Usage:
            write_reg <cmd_name> <data> [offset] [timeout]

        Args:
            cmd_name: The name of the register to write
            data: The data to write to the register
            offset: The number of elements to offset in an array
            timeout: Optional timeout value for command specific timeouts

        Example:
            To write to user register 0 the data 42 we can do this
            write_reg user_reg 42 0

            To write many bytes the data must not be seperated by spaces
            write_reg user_reg [1,2,3,4,5,6] 0
        """
        self._print_func_result(self.phil.write_reg, arg)

    def complete_write_reg(self, text, line, begidx, endidx):
        """Completes arg with memory map record names"""
        begidx = begidx
        endidx = endidx
        return self._complete_map(text, line)

    def do_read_struct(self, arg):
        """Reads a set of registers defined by the memory map

        Usage:
            read_struct <cmd_name> [timeout]

        Args:
            cmd_name: The name if the structure to read
            timeout: Optional timeout value for command specific timeouts
        """
        self._print_func_result(self.phil.read_struct, arg)

    def complete_read_struct(self, text, line, begidx, endidx):
        """Completes arg with memory map record structs"""
        begidx = begidx
        endidx = endidx
        mline = line.partition(' ')[2]
        offs = len(mline) - len(text)
        map_records = [*self.phil.mem_map]
        map_structs = []
        for record in map_records:
            first_name = record.split('.')[0]
            if first_name not in map_structs:
                map_structs.append(first_name)
        return [s[offs:] for s in map_structs if s.startswith(mline)]

    def do_execute_changes(self, arg):
        """Executes or commits device configuration changes

        This will cause any changes in configuration to be applied. For many
        periphs the .mode.init must be set to 0 for the periph to reinitialize.

        Usage:
            execute_changes [timeout]

        Args:
            timeout: Optional timeout value for command specific timeouts
        """
        self._print_func_result(self.phil.execute_changes, arg)

    def do_philip_reset(self, arg):
        """Resets the device

        Usage:
            philip_reset [timeout]

        Args:
            timeout: Optional timeout value for command specific timeouts
        """
        self._print_func_result(self.phil.reset_mcu, arg)

    def do_get_version(self, arg):
        """Get the version of the interface/memory map that is being used

        Usage:
            get_version
        """
        arg = arg
        print(self.phil.if_version)

    def do_write_and_execute(self, arg):
        """Writes the register and the init for the struct and executes changes

        Usage:
            write_and_execute <cmd_name> <data> [timeout]

        Args:
            cmd_name: The name of the register to write
            data: The data to write to the register
            timeout: Optional timeout value for command specific timeouts
        """
        self._print_func_result(self.phil.write_and_execute, arg)

    def complete_write_and_execute(self, text, line, begidx, endidx):
        """Completes arg with memory map record names"""
        begidx = begidx
        endidx = endidx
        return self._complete_map(text, line)

    def do_dut_reset(self, arg):
        """Provides a reset to the dut

        Toggles the dut reset pin for the reset time.

        Usage:
            dut_reset [reset_time] [timeout]

        Args:
            reset_time: Optional duration the dut is put in reset
            timeout: Optional timeout value for command specific timeouts
        """
        self._print_func_result(self.phil.dut_reset, arg)

    def do_read_trace(self, arg):
        """Reads event trace from the dut

        Usage:
            read_trace
        """
        self._print_func_result(self.phil.read_trace, arg)

    def do_data_filter(self, arg):
        """Select or toggle filtering for data

        Usage:
            data_filter [{ON, OFF}]

        Args:
            {ON, OFF}: Turn filtering on or off, if no arg it toggles
        """
        if arg:
            if arg.upper() == "ON":
                self.data_only = True
                print("Filtering for data")
            elif arg.upper() == "OFF":
                self.data_only = False
                print("Raw data, no filtering")
            else:
                print("Incorrect arg")
        elif self.data_only:
            self.data_only = False
            print("Raw data, no filtering")
        else:
            self.data_only = True
            print("Filtering for data")

    def do_print_map(self, arg):
        """Prints the device map

        Usage:
            print_map [cmd_name]

        Args:
            cmd_name: The name of the register in the map to print
        """
        arg = arg
        if arg:
            try:
                print(dumps(self.phil.mem_map[arg], sort_keys=True, indent=4))
            except KeyError as exc:
                print("Cannot parse {}".format(exc))
        else:
            print(dumps(self.phil.mem_map, sort_keys=True, indent=4))

    def complete_print_map(self, text, line, begidx, endidx):
        """Completes arg with memory map record names"""
        begidx = begidx
        endidx = endidx
        return self._complete_map(text, line)

    def _complete_map(self, text, line):
        mline = line.partition(' ')[2]
        offs = len(mline) - len(text)
        map_records = [*self.phil.mem_map]
        return [s[offs:] for s in map_records if s.startswith(mline)]

    def do_info_record_type(self, arg):
        """Prints the device map

        Usage:
            info_record_type <record_type>

        Args:
            record_type: The record type in a map, such as description
        """
        try:
            record_types = {}
            for key, val in self.phil.mem_map.items():
                if arg in val:
                    if val[arg]:
                        record_types[key] = val[arg]
            print(dumps(record_types, sort_keys=True, indent=4))
        except KeyError as exc:
            print("Cannot parse {}".format(exc))

# pylint: disable=R0201
    def complete_info_record_type(self, text, line, begidx, endidx):
        """Completes arg with common record types"""
        begidx = begidx
        endidx = endidx
        mline = line.partition(' ')[2]
        offs = len(mline) - len(text)
        info_record_types = ['description', 'access', 'default', 'bit',
                             'flag', 'max', 'min']
        return [s[offs:] for s in info_record_types if s.startswith(mline)]

    def do_show_pinout(self, arg):
        """Prints the pinout for the connected board

        Usage:
            show_pinout
        """
        try:
            if arg:
                showboard = int(arg)
            else:
                showboard = self.phil.read_reg('sys.status.board')['data']
            if showboard == 1:
                print("""
PHILIP-B -> BLUEPILL

                    ____
                 ___|__|___
 DUT_RST = B12 - |        | - GND
 DUT_CTS = B13 - |        | - GND
 DUT_RTS = B14 - |        | - 3V3
USER_BTN = B15 - |        | - NRST
   DUT_IC = A8 - |        | - B11 = DUT_RX
    IF_TX = A9 - |        | - B10 = DUT_TX
   IF_RX = A10 - |        | - B1 = PM_V_ADC
  USB_DM = A11 - |        | - B0 = PM_HI_ADC
  USB_DP = A12 - |        | - A7 = PM_LO_ADC
 DUT_NSS = A15 - |        | - A6 = DUT_ADC
  DUT_SCK = B3 - |        | - A5 = TEST_FAIL
 DUT_MISO = B4 - |        | - A4 = TEST_WARN
 DUT_MOSI = B5 - |        | - A3 = TEST_PASS
  DUT_SCL = B6 - |        | - A2 = DEBUG2
  DUT_SDA = B7 - |        | - A1 = DEBUG1
  DUT_PWM = B8 - |        | - A0 = DEBUG0
  DUT_DAC = B9 - |        | - C15
            5V - |        | - C14
           GND - |        | - C13 = LED0
           3V3 - |        | - VBAT
                 __________
                    ||||
""")
            else:
                print("""
PHILIP-N -> NUCLEO-F103RB
CN6

                            DUT_SCL = PB8 = SCL/D15 -
                            DUT_SDA = PB9 = SDA/D14 -
                                               AVDD -
                                                GND -
-                               LED0 = PA5 =SCK/D13 -
- IOREF                                    MISO/D12 -
- NRST                                 PWM/MOSI/D11 -
- 3V3                                    PWM/CS/D10 -
- 5V                                         PWM/D9 -
- GND                             DUT_TX = PA9 = D8 -
- GND                                           |CN5|
- VIN                             DUT_IC = PA8 = D7 -
|CN6|                                        PWM/D6 -
- A0 = PA0 = TEST_WARN        DEBUG1 = PB4 = PWM/D5 -
- A1 = PA1 = TEST_FAIL            DEBUG2 = PB5 = D4 -
- A2 = PA4 = TEST_PASS        DEBUG0 = PB3 = PWM/D3 -
- A3 = PB0 = DUT_ADC             DUT_RX = PA10 = D2 -
- A4 = PC1 = PM_HI_ADC          IF_TX = PA2 = TX/D1 -
- A5 = PC0 = PM_V_ADC           IF_RX = PA3 = RX/D0 -
|CN8|                                          |CN10|

          -1 -                  DUT_DAC -1 - DUT_PWM
          -2 -                  DUT_SCL -2 -
          -3 -                  DUT_SDA -3 -
          -4 -                          -4 -
          -5 -                          -5 -
          -6 -                     LED0 -6 - DUT_RTS
          -7 -                          -7 - DUT_CTS
          -8 -                          -8 - DUT_NSS
          -9 -                          -9 -
          -10-                          -10-
          -11-                   DUT_TX -11- DUT_RST
 USER_BTN -12-                   DUT_IC -12-
          -13-                          -13- DUT_MOSI
          -14- TEST_WARN         DEBUG1 -14- DUT_MISO
          -15- TEST_FAIL         DEBUG2 -15- DUT_SCK
          -16- TEST_PASS         DEBUG0 -16-
          -17- DUT_ADC           DUT_RX -17-
PM_LO_ADC -18- PM_HI_ADC          IF_TX -18-
          -19- PM_V_ADC           IF_RX -19-
          |CN7|                         |CN10|
""")
        except (ValueError) as exc:
            print(exc)

    def do_exit(self, arg):
        """I mean it should be obvious

        Usage:
            exit
        """
        arg = arg
        return True

    def _print_func_result_success(self, results):
        if not isinstance(results, list):
            results = [results]
        result = self.phil.RESULT_SUCCESS
        printed_something = False
        for res in results:
            if self.data_only:
                if 'result' in res:
                    if res['result'] is self.phil.RESULT_SUCCESS:
                        if 'data' in res:
                            print(dumps(res['data']))
                            printed_something = True
                    else:
                        result = res['result']
                else:
                    print(dumps(res))
                    printed_something = True
            else:
                print(dumps(res))
                printed_something = True
        if not printed_something:
            print(result)

    def _print_func_result(self, func, arg):
        value = (arg or '').split(' ')
        func_args = [v for v in value if v]
        try:
            results = func(*func_args)
        except KeyError as exc:
            print('Could not parse argument {}'.format(exc))
        except (TypeError, ValueError, SyntaxError) as exc:
            print(exc)
        else:
            self._print_func_result_success(results)


PARSER = argparse.ArgumentParser()

LOG_LEVELS = ('debug', 'info', 'warning', 'error', 'fatal', 'critical')
PARSER.add_argument('--loglevel', choices=LOG_LEVELS, default='info',
                    help='Python logger log level')
PARSER.add_argument('--port', '-p', help='Specifies the serial port',
                    default=None)
PARSER.add_argument('--data_filter_off', '-do', default=True,
                    action='store_false',
                    help='Filters data from philip responses')
PARSER.add_argument('--use_dev_map', '-dm', default=False,
                    action='store_true',
                    help='Uses the memory map from device')


def _exit_cmd_loop():
    if readline:
        try:
            readline.write_history_file()
        except IOError:
            pass


def main():
    """Main program"""
    pargs = PARSER.parse_args()

    if pargs.loglevel:
        loglevel = logging.getLevelName(pargs.loglevel.upper())
        logging.basicConfig(level=loglevel)
    try:
        PhilipShell(port=pargs.port, data_only=pargs.data_filter_off,
                    use_dev_map=pargs.use_dev_map).cmdloop()
        _exit_cmd_loop()
    except KeyboardInterrupt:
        _exit_cmd_loop()


if __name__ == '__main__':
    main()
